;**********************************************************************
;                                                                     *
;    Filename:	    sio.asm                                           *
;    Date:          5 April 2003                                      *
;    File Version:                                                    *
;                                                                     *
;    Author: Al Williams                                              *
;    Company: AWC                                                     *
;                                                                     * 
;                                                                     *
;**********************************************************************
;                                                                     *
;    Notes: This is a simple I/O system made to be driven             *
;           by a PC or similar host. You send a byte:                 *
;               CCCCPPPP                                              *
;           Where CCCC is a command and PPPP is a pin number          *
;           The commands recognized are:                              *
;             0 - Set pin low                                         *
;             1 - Set pin high                                        *
;             2 - Read pin (returns "0" or "1" in ASCII)              *
;             3 - Toggle pin (high to low, or low to high)            *
;    Commands that set outputs (0,1,3) set the pin to output          *
;    Pins are inputs by default, and command 2 resets the             *
;    state to input. See code for mapping of PPPP                     *
;                                                                     *
;**********************************************************************


	list      p=16f873            ; list directive to define processor
	#include <p16f873.inc>        ; processor specific variable definitions
	
	__CONFIG _CP_OFF & _WDT_ON & _BODEN_ON & _PWRTE_ON & _RC_OSC & _WRT_ENABLE_ON & _LVP_ON & _CPD_OFF



; 19200 baud
BAUD_CONSTANT	EQU	0x40	;Constant for baud generator for 19200@20Mhz


; I am not using interrupts, but there is plenty of room,
; so I left the ISR underpinning in place for future
; expansion
;***** VARIABLE DEFINITIONS
w_temp        EQU     0x20        ; variable used for context saving 
status_temp   EQU     0x21        ; variable used for context saving

		cblock 0x22
cmdbyte:	1
maska:		1	   
maskb:		1
maskc:		1
ashadow:	1       ; be sure to reserve Xshadow in bank 1 too!
bshadow:	1
cshadow:	1
temp:		1
		endc







;**********************************************************************
		ORG     0x000             ; processor reset vector
		clrf    PCLATH            ; ensure page bits are cleared
  		goto    main              ; go to beginning of program


		ORG     0x004             ; interrupt vector location
		movwf   w_temp            ; save off current W register contents
		movf	STATUS,w          ; move status register into W register
		bcf     STATUS,RP0        ; ensure file register bank set to 0
		movwf	status_temp       ; save off contents of STATUS register


; isr code can go here or be located as a call subroutine elsewhere


		bcf     STATUS,RP0        ; ensure file register bank set to 0
		movf    status_temp,w     ; retrieve copy of STATUS register
		movwf	STATUS            ; restore pre-isr STATUS register contents
		swapf   w_temp,f
		swapf   w_temp,w          ; restore pre-isr W register contents
		retfie                    ; return from interrupt


; Here's the mainline stuff
; I keep a shadow of PORTA, PORTB, and PORTC in banks 0 and 1
; (bank 1 copies are really TRISA, TRISB, and TRISC)
; This end runs the read/modify/write problem!
main:

		clrf	ashadow   ; all outputs are 0
		clrf	bshadow
		clrf	cshadow
		movlw   0xFF      ; and all pins are inputs
		bsf 	status,rp0
		movwf	ashadow
		movwf	bshadow
		movwf	cshadow
		movlw	6	; turn off A/D converter
		movwf	adcon1
		bcf	status,rp0
		call	SerialSetup  ; init UART
mainloop:
		call	SerialReceive  ; wait for command
; The command structure:
; CCCCPPPP
; where CCCC = Command
;       PPPP = Port (0-15)
; Commands:
; 0 - Set low (forces output)
; 1 - Set high (forces output)
; 2 - Read (forces input)
; 3 - Toggle (forces output)
;
; Port mapping:
; 0-5 = PortC 0-5
; 6-7 = PortB 0-1
; 8-13 = PORTA 0-5 (note PORT A4 is open collector/schmitt)
; 14-15 = PORTB 4-5
		movwf	cmdbyte
		call	setmask ; set masks for porta/b/c
		swapf	cmdbyte,w
		andlw	0xf    ; strip off port info
; jump table to correct command (plenty of extra room :-) )
		addwf	pcl,f
		goto    clow
		goto	chigh
		goto	cread
		goto	ctoggle
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop
		goto	mainloop

; command to set low pin
clow:
; the idea is to invert the masks and logical and them with
; the corresponding port shadow. We also and them with the TRIS shadows
; to make the pin an output. The done routine (below)
; actually moves the shadows to their destination.
	comf maska,w
	andwf ashadow,f
	bsf status,rp0
	andwf ashadow,f
	bcf status,rp0
	comf maskb,w
	andwf bshadow,f
	bsf status,rp0
	andwf bshadow,f
	bcf status,rp0
	comf maskc,w
	andwf cshadow,f
	bsf status,rp0
	andwf cshadow,f
	bcf status,rp0

; This routine sets all the shadow registers to their physical
; counterparts.
done:
	movf ashadow,w
	movwf porta
	movf bshadow,w
	movwf portb
	movf cshadow,w
	movwf portc
	call setdirs
	goto mainloop

setdirs:
; set data direction registers to match shadows
	bsf status,rp0
	movf ashadow,w
	movwf porta
	movf bshadow,w
	movwf portb
	movf cshadow,w
	movwf portc
	bcf status,rp0
	return

; Set a pin high
; This uses the same logic as low except it doesn't invert
; the masks and ORs it with the data shadow. Then it does
; invert the mask and ands it with the TRIS shadow so the
; pin is forced to output
chigh:
	movf maska,w
	iorwf ashadow,f
	xorlw 0xff
	bsf status,rp0
	andwf ashadow,f
	bcf status,rp0
	movf maskb,w
	iorwf bshadow,f
	xorlw 0xff
	bsf status,rp0
	andwf bshadow,f
	bcf status,rp0
	movf maskc,w
	iorwf cshadow,f
	xorlw 0xff
	bsf status,rp0
	andwf cshadow,f
	bcf status,rp0
	goto done

; Reading is different because we have to change the 
; direction register first. Then we just
; read (A & maska) | (B & maskb) | (C & maskc)
; and generate an ASCII 0 or 1 accordingly
cread:
	movf maska,w
	bsf status,rp0
	iorwf ashadow,f
	bcf status,rp0
	movf maskb,w
	bsf status,rp0
	iorwf bshadow,f
	bcf status,rp0
	movf maskc,w
	bsf status,rp0
	iorwf cshadow,f
;	bcf status,rp0  -- setdirs will clear this for us!
	call setdirs
; now the bit is an input so read the three ports and mask them
	movf porta,w
	andwf maska,w
	movwf temp
	movf portb,w
	andwf maskb,w
	iorwf temp,f
	movf portc,w
	andwf maskc,w
	iorwf temp,f
	movlw "0"        ; assume 0
	btfss status,z
	addlw 1          ; but if not, make it a "1"
	call SerialTransmit  ; send it
	goto mainloop

; toggle is just like high except it xors the masks instead of
; or'ing them
ctoggle:
	movf maska,w
	xorwf ashadow,f
	xorlw 0xff
	bsf status,rp0
	andwf ashadow,f
	bcf status,rp0
	movf maskb,w
	xorwf bshadow,f
	xorlw 0xff
	bsf status,rp0
	andwf bshadow,f
	bcf status,rp0
	movf maskc,w
	xorwf cshadow,f
	xorlw 0xff
	bsf status,rp0
	andwf cshadow,f
	bcf status,rp0
	goto done

; this subroutine converts the pin # in cmdbyte(3:0)
; to a single bit in the 24 bits of maskc, maskb, and maska
; since the arrangement is sort of arbitrary
; the code just divides up into slices by a look up
; table
setmask:
	clrf maska
	clrf maskb
	clrf maskc
	movlw 0xF
	andwf cmdbyte,w
	addwf pcl,f
	goto pin05
	goto pin05
	goto pin05
	goto pin05
	goto pin05
	goto pin05
	goto pin67
	goto pin67
	goto pin813
	goto pin813
	goto pin813
	goto pin813
	goto pin813
	goto pin813
	goto pin1415
	goto pin1415

; These are the four distinct "slices"
; the tobit subroutine just converts 0 => 1, 1 => 10, 2 => 100, etc.
pin05:
	call tobit  ; this slice is easy because they are in order in portc
	movwf maskc
	return

pin67:
	addlw -6    ; pins 6 and 7 are really maskb 0 and 1!
	call tobit
	movwf maskb
	return

pin813:
	addlw -8   ; 8-13 are really 0-5 in maska
	call tobit
	movwf maska
	return

pin1415:
	addlw -.10   ; 14 and 15 are really 4 and 5 in maskb
	call tobit
	movwf maskb
	return

; just a lookup table to compute 2**w
tobit:
	addwf pcl,f
	retlw 1
	retlw 2
	retlw 4
	retlw 8
	retlw 0x10
	retlw 0x20
	retlw 0x40
	retlw 0x80

; hardware uart set up - right out of the manual
SerialSetup:	
		bsf     STATUS,RP0
		movlw	BAUD_CONSTANT	;set baud rate 
		movwf	SPBRG
		bsf	TXSTA,BRGH	;baud rate high speed option
		bsf	TXSTA,TXEN	;enable transmission
		bcf     STATUS,RP0
		bsf	RCSTA,CREN	;enable reception
		bsf	RCSTA,SPEN	;enable serial port
		bcf	STATUS,RP0
		return

;-----------------------------------------------------------------------------
;Wait for byte to be received in UART and return with byte in W

SerialReceive:	
		btfss	PIR1,RCIF	;check if data received
		goto	$-1		;wait until new data
		movf	RCREG,W		;get received data into W
		return


;-----------------------------------------------------------------------------
;Transmit byte in W register from USART

SerialTransmit:	
		btfss	PIR1,TXIF	;check that buffer is empty
		goto	$-1
		movwf	TXREG		;transmit byte
		return



		END                       

