package com.mindsnack.test;

import junit.framework.TestCase;

import com.mindsnack.transactions.Copier;
import com.mindsnack.transactions.PlainCopier;
import com.mindsnack.transactions.SimpleTransaction;
import com.mindsnack.transactions.Transaction;
import com.mindsnack.transactions.TransactionNotBegunException;

import com.mindsnack.triangles.*;

public class TestTriangle extends TestCase
{

  public void testSetsMustBeInTransactions()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t = new TriangleImpl(60.0,60.0,60.0);

    Triangle triangle = (Triangle) transaction.addToTransaction(t, new PlainCopier());
    t.setAngle(1,60.0); // should be OK, t is not a proxy.
    try {
      triangle.setAngle(1,60.0);
      fail("Should not be able to 'set' if the transaction has not begun.");
    } catch (TransactionNotBegunException e)
      { }
    transaction.begin();
    {
      try {
        triangle.setAngle(1,60.0);
      } catch (TransactionNotBegunException e)
        { fail("Setting angle within a transaction should be OK."); }
    }
    transaction.commit();
  }

  public void testGetsDontHaveToBeInTransactions()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t = new TriangleImpl(60.0,60.0,60.0);

    Triangle triangle = (Triangle) transaction.addToTransaction(t, new TriangleCopier());
    t.getAngle(1); // should be OK.
    try {
      triangle.getAngle(1);
    } catch (TransactionNotBegunException e)
      { fail("Get's should be OK outside of transaction."); }

    transaction.begin();
      triangle.setAngle(1,60.0);
    transaction.commit();

    try {
      triangle.getAngle(1);
    } catch (TransactionNotBegunException e)
      { fail("Get's should be OK outside of transaction."); }

  }

  public void testSuccessfulTransaction()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t = new TriangleImpl(60.0,60.0,60.0);

    Triangle triangle = (Triangle) transaction.addToTransaction(t, new TriangleCopier());

    transaction.begin();
    triangle.setAngle(1, 100.0);
    triangle.setAngle(2, 40.0);
    triangle.setAngle(3, 40.0);
    assertTrue(t.getAngle(1) == 60.0);  // haven't committed the value yet.
    assertTrue(triangle.getAngle(1) == 100.0);
    boolean result = transaction.commit();
    assertTrue(result);
    assertTrue(t.getAngle(1) == 100.0);
    assertTrue(triangle.getAngle(1) == 100.0);
  }

  public void testInvalidTransaction()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t = new TriangleImpl(60.0,60.0,60.0);

    Triangle triangle = (Triangle) transaction.addToTransaction(t, new TriangleCopier());

    transaction.begin();
    triangle.setAngle(1, 100.0);
    triangle.setAngle(2, 40.0);
    assertTrue(t.getAngle(1) == 60.0);  // haven't committed the value yet.
    assertTrue(triangle.getAngle(1) == 100.0);

    boolean result = transaction.commit();  // Transaction should fail.
    assertTrue(result==false);
    assertTrue(t.getAngle(1) == 60.0);
    assertTrue(triangle.getAngle(1) == 60.0);
  }

  public void testRolledBackTransaction()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t = new TriangleImpl(60.0,60.0,60.0);

    Triangle triangle = (Triangle) transaction.addToTransaction(t, new TriangleCopier());

    transaction.begin();
    triangle.setAngle(1, 100.0);
    triangle.setAngle(2, 40.0);
    triangle.setAngle(3, 40.0);
    transaction.rollback();
    assertTrue(t.getAngle(1) == 60.0);
    assertTrue(triangle.getAngle(1) == 60.0);
  }

  public void testInvalidTransaction2Triangles()
  {
    Transaction transaction = new SimpleTransaction();

    Triangle t1 = new TriangleImpl(10.0,10.0,160.0);
    Triangle t2 = new TriangleImpl(20.0,20.0,140.0);

    Triangle triangle1 = (Triangle) transaction.addToTransaction(t1, new TriangleCopier());
    Triangle triangle2 = (Triangle) transaction.addToTransaction(t2, new TriangleCopier());

    transaction.begin();
    triangle1.setAngle(1, 100.0);
    triangle1.setAngle(2, 40.0);
    triangle1.setAngle(3, 40.0);

    triangle2.setAngle(1, 60.0);
    triangle2.setAngle(2, 60.0);

    boolean result = transaction.commit();  // Transaction should fail.
    assertTrue(result==false);
    assertTrue(t1.getAngle(1) == 10.0);  // because t2 is invalid.
    assertTrue(triangle1.getAngle(1) == 10.0);
    assertTrue(t2.getAngle(1) == 20.0);
    assertTrue(triangle2.getAngle(1) == 20.0);
  }
}
