package com.mindsnack.transactions;

import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class SimpleTransaction
  implements Transaction
{
  protected List list = new ArrayList();
  protected boolean transactionStarted = false;
  protected boolean rollbackOnly = false;


  public boolean isBegun()
    { return transactionStarted; }

  public void setRollbackOnly()
    { rollbackOnly = true; }

  public boolean isRollbackOnly()
    { return rollbackOnly; }


  protected void checkInTransaction()
  {
    if (!isBegun())
      throw new TransactionNotBegunException();
  }

  // Helper methods.
  protected synchronized void actualStartTransaction()
    { transactionStarted = true; rollbackOnly = false; }

  protected synchronized void actualReleaseTransaction()
    { transactionStarted = false; }
  // End Helper Methods.

  public synchronized void begin()
  {
    if (transactionStarted)
      throw new TransactionBegunException();


    // Let TransactionListener know that a transaction is beginning.
    Iterator iter = list.iterator();
    while (iter.hasNext())
    {
      TransactionProxy proxy = (TransactionProxy) iter.next();
      if (proxy.duplicate instanceof TransactionListener)
      {
        TransactionListener transactionListener = (TransactionListener) proxy.duplicate;
        transactionListener.begin(this);
      }
    }

    transactionStarted = true;
    rollbackOnly = false;
  }

  public synchronized void rollback()
  {
    checkInTransaction();

    // Let TransactionListener do a rollback if they want to.
    Iterator iter = list.iterator();
    while (iter.hasNext())
    {
      TransactionProxy proxy = (TransactionProxy) iter.next();
      if (proxy.duplicate instanceof TransactionListener)
      {
        TransactionListener transactionListener = (TransactionListener) proxy.duplicate;
        transactionListener.rollback(this);
      }
    }
    actualReleaseTransaction();
  }


  public synchronized boolean commit()
  {
    // Two phase commit.
    checkInTransaction();

    // phase 1 - ask if every object is OK with a commit
    Iterator iter = list.iterator();
    while (iter.hasNext() && !isRollbackOnly())
    {
      TransactionProxy proxy = (TransactionProxy) iter.next();
      if (proxy.duplicate instanceof Validatable)
      {
        Validatable validatable = (Validatable) proxy.duplicate;
        if (!validatable.isValid())
        {
          setRollbackOnly();
          break;
        }
      }

      if (proxy.duplicate instanceof TransactionListener)
      {
        TransactionListener transactionListener = (TransactionListener) proxy.duplicate;
        if (!transactionListener.agreeToCommit(this))
        {
          setRollbackOnly();
          break;
        }
      }
    }

    // Phase 2 - Perform the commit
    if (!isRollbackOnly())
    {
      iter = list.iterator();
      while (iter.hasNext())
      {
        TransactionProxy proxy = (TransactionProxy) iter.next();
        if (proxy.duplicate instanceof TransactionListener)
        {
          TransactionListener transactionListener = (TransactionListener) proxy.duplicate;
          transactionListener.commit(this);
        }

        Copier copier = proxy.copier;
        copier.copy(proxy.duplicate, proxy.original);
      }
    } else {  // doing a rollback
      rollback();
    }

    boolean result = !isRollbackOnly();
    actualReleaseTransaction();
    return result;
  }


  public Object addToTransaction(Object original, Copier copier)
  {
    TransactionProxy proxy = null;

    try {
      proxy = new TransactionProxy(original, copier, this);
    } catch (InstantiationException ignore) { return null; }
      catch (IllegalAccessException ignore) { return null; }

    synchronized(list) {
      list.add(proxy);
    }

    Object dynamicProxy = Proxy.newProxyInstance(
             original.getClass().getClassLoader(),
             original.getClass().getInterfaces(),
             proxy
             );

    return dynamicProxy;
  }
}
