package com.ericgiguere;

import java.awt.*;
import java.awt.event.*;
import javax.microedition.xlet.*;

/**
 * Defines a class that lets you run an Xlet
 * using the traditional application model.  As a side benefit,
 * it also lets you run the Xlet on J2SE platforms, providing
 * you have all the necessary classes in the classpath.
 */

public class XletRunner {

    // State definitions, including private states
    // for internal use only.

    private static final int STATE_CREATING = 0;
    private static final int STATE_LOADED = 1;
    private static final int STATE_PAUSED = 2;
    private static final int STATE_ACTIVE = 3;
    private static final int STATE_DESTROYED = 4;
    private static final int STATE_ACTIVATION_REQUESTED = 5;
    private static final int STATE_DESTRUCTION_REQUESTED = 6;

    // Hold a reference to the Xlet and to its current state.

    private Xlet  xlet = null;
    private int   state = STATE_CREATING;
    private Frame frame = null;

    // Prints an error message and then quits.

    protected void abortXlet( String msg, Throwable e ){
        System.err.println( msg );
        e.printStackTrace( System.err );
        System.exit( 1 );
    }

    // One way to run an Xlet: call this class and
    // pass the Xlet's class name as the first parameter.
    // The Xlet is invoked and the remaining arguments are
    // passed to it.

    public static void main( String[] args ){
        if( args.length == 0 ){
            System.err.println( "Usage: java com.ericgiguere.XletRunner [classname] <args>" );
            System.exit( 1 );
        }

        String[] xletArgs = new String[ args.length-1 ];
        System.arraycopy( args, 1, xletArgs, 0, args.length-1 );

        new XletRunner().runXlet( args[0], xletArgs );
    }

    // Run an Xlet. Creates a new instance of the Xlet and
    // activates it according to the PBP spec.

    public void runXlet( Class cl, String[] args ){
        XletContext context = null;
        Throwable   exception = null;

        trace( "Loading Xlet of class "+ cl.getName() );

        // Create the Xlet

        try {
            xlet = (Xlet) cl.newInstance();
        }
        catch( Throwable e ){
            abortXlet( "Xlet " + cl.getName() +
                       " could not be instantiated", e );
        }

        // Place it in the loaded state and call initXlet

        trace( "Xlet is in LOADED state, calling initXlet" );

        context = new Context( args );
        state = STATE_LOADED;

        try {
            xlet.initXlet( context );
            state = STATE_ACTIVATION_REQUESTED;

            trace( "Xlet is initialized, in PAUSED stated, " +
                   "requesting activation" );
        }
        catch( Throwable e ){
            exception = e;
        }

        // If initXlet did not throw any exceptions, we can
        // start the Xlet.  Otherwise we immediately destroy it.
        // Basically, we just loop around waiting for state
        // changes.

        while( exception == null ){
            if( state == STATE_ACTIVATION_REQUESTED ){
                trace( "Activating Xlet by calling startXlet" );

                try {
                    state = STATE_ACTIVE;
                    frame.setVisible( true );
                    xlet.startXlet();

                    trace( "Xlet is in the ACTIVE state" );
                }
                catch( Throwable e ){
                    exception = e;
                }
            } else if( state == STATE_DESTRUCTION_REQUESTED ){
                trace( "Conditionally destroying Xlet by calling " +
                       "destroyXlet( false )" );

                try {
                    xlet.destroyXlet( false );
                    state = STATE_DESTROYED;
                }
                catch( XletStateChangeException xsce ){
                    trace( "Destruction aborted by destroyXlet" );
                    state = STATE_ACTIVE; // abort destruction
                }
                catch( Throwable e ){
                    exception = e;
                }
            } else if( state == STATE_PAUSED ){
                if( frame.isVisible() ){
                    frame.setVisible( false );
                }
            }

            if( state == STATE_DESTROYED ) break;
            if( exception != null ) break;

            synchronized( context ){
                try {
                    context.wait();
                }
                catch( InterruptedException e ){
                }
            }
        }

        // The Xlet is terminating.  If it is not already in
        // the destroyed state, call destroyXlet.

        if( state != STATE_DESTROYED ){
            try {
                trace( "Xlet entering DESTROYED state, calling " +
                       "destroyXlet( true )" );
                xlet.destroyXlet( true );
            }
            catch( Throwable e ){
                // ignore any exceptions here
            }
        }

        if( exception != null ){
            abortXlet( "Xlet terminated due to uncaught exception",
                       exception );
        }

        trace( "Xlet is in DESTROYED state and has terminated" );
	frame.dispose();
        System.exit( 0 );
    }

    // Run an xlet given a classname and some arguments.

    public void runXlet( String classname, String[] args ){
        try {
            runXlet( Class.forName( classname ), args );
        }
        catch( ClassNotFoundException e ){
            abortXlet( "Class " + classname + " was not found", e );
        }
    }

    // A simple tracing function.

    protected void trace( String msg ){
        System.out.println( "[XletRunner] " + msg );
    }

    /**
     * An inner class defining the Xlet's context.  A
     * simple enough class, it mostly just changes the
     * state of the Xlet and then wakes up the main thread.
     */

    private class Context extends WindowAdapter
                          implements XletContext {
        private String[] args;
        
        Context( String[] args ){
            this.args = args;

            Dimension size = Toolkit.getDefaultToolkit().getScreenSize();
            frame = new Frame();
            frame.setBounds( 0, 0, size.width, size.height );
            frame.addWindowListener( this );
        }
        
        public Container getContainer()
                     throws UnavailableContainerException {
            return frame;
        }

        public Object getXletProperty( String key ){
            if( key == ARGS ){
                return args;
            }

            return null;
        }

        public synchronized void notifyDestroyed(){
            state = STATE_DESTROYED;
            this.notify();
        }

        public synchronized void notifyPaused(){
            if( state == STATE_ACTIVE ||
                state == STATE_ACTIVATION_REQUESTED ){
                state = STATE_PAUSED;
                this.notify();
            }
        }

        public synchronized void resumeRequest(){
            if( state == STATE_PAUSED ){
                state = STATE_ACTIVATION_REQUESTED;
                this.notify();
            }
        }

        public synchronized void windowClosing( WindowEvent e ){
            if( state == STATE_ACTIVE ){
                state = STATE_DESTRUCTION_REQUESTED;
                this.notify();
            }
        }
    }
}
